/* 
 * KeyHole v0.1 beta 3 written by Asynchro & Cthulhu (pk members)
 *   
 * A tool for give to each user the rights to connect to some ports 
 * through a firewall so the admin can close even the port that have
 * to be used remotely 
 *
 * Config_file syntax:
 *
 * Username:portlist: 
 *  portlist can be a coma separated list of ports or ports range 
 *  ex: root:21,23,65-78,8080:  
 *
*/

#include "keyhole.h"


char *plist;
char portlist[8192];
char filename[MAXBUFF];
char ifname[IFNAMSIZ];

ip_chainlabel chain = "input";
int a, maxconnum = MAXCON;


struct sockaddr_in saddr, caddr;
struct saved_ipfw *first_ipfw, *actual_ipfw;


struct cmd cmdlist[] = {

	  {"LIST", list,
		    "LIST                  shows all avaliable ports\n"},
	  {"OPEN", fwopen,
		    "OPEN <portnum>        opens a specified port\n"},
	  {"QUIT", quit,
		    "QUIT                  closes connection\n"},
	  {"CLOSE", fwclose,
		    "CLOSE <portnum>       closes a specified port\n"},
	  {"SHOW", fwshow,           
                    "SHOW                  shows all the opened ports\n"},
          {"HELP", help,
		    "HELP                  shows this brief help message\n"},
	  {NULL, NULL, NULL}
};




main (int argc, char **argv)
{

	  struct sigaction act;
	  int opt, port;
	  char debug = 0;

	  port = PORT;
	  strncpy (filename, CONFIG_FILE, MAXBUFF);
          strncpy (ifname, INTERFACE, MAXBUFF);

          if(getuid())
          {fprintf(stderr,"You have to be root\nExiting\n"); exit(ERR);}

	  while ((opt = getopt (argc, argv, "dhm:p:f:i:")) != EOF)
	    {
		      switch (opt)
			{
			case 'd':
				  /*
				     don't demonize 
				   */
				  debug++;
				  break;
			case 'm':
				  maxconnum = atoi (optarg);

                        case 'i':
                                  strncpy(ifname, optarg, IFNAMSIZ);                                               break;
			case 'h':
				  usage (argv[0]);
				  break;
			case 'p':
				  port = atoi (optarg);
				  break;
			case 'f':
				  /*
				     configuration's file name 
				   */
				  strncpy (filename, optarg, MAXBUFF);
				  break;
			case '?':
			default:
				  usage (argv[0]);
				  break;
			}
	    }


	  if (!debug)
		    demonize ();

	  syslog (LOG_INFO | LOG_USER, "rfw started on port %d\n", port);
	  sigemptyset (&act.sa_mask);
	  act.sa_flags = 0;

	  act.sa_handler = SIG_IGN;
	  sigaction (SIGCHLD, &act, 0);
	  act.sa_handler = endserver;
	  sigaction (SIGINT, &act, 0);
	  sigaction (SIGQUIT, &act, 0);

	  first_ipfw = malloc (sizeof (struct saved_ipfw));
	  actual_ipfw = first_ipfw;


	  memset (portlist, 0x0, 8192);

	  server (port);

}


void
usage (char *argv)
{
	  printf
		    ("\nKey Hole v0.1 beta 2\nby Asynchro & Cthulhu (pk crew)\n\nUsage: %s [-p port] [-f config_file] [-m max connection number] [-i interface] [-d] [-h]\n\n",
		    argv);
	  exit (OK);
}


int
demonize (void)
{
	  int fd;
	  int ppid;
	  close (0);
	  close (1);
	  close (2);
	  umask (0);
	  setpgrp ();

	  if ((fd = open ("/dev/tty", O_RDWR)) >= 0)
	    {
		      ioctl (fd, TIOCNOTTY, NULL);
		      close (fd);
	    }
	  if (ppid = fork () < 0)
	    {
		      perror ("fork:");
		      exit (ERR);
	    }
	  else if (ppid)
		    exit (OK);
	  return (OK);
}


/* endserver: close the main server and send SIGQUIT to all children process */

void
endserver (int sig)
{
	  kill (0, SIGQUIT);

	  fprintf (stderr, "Quitting\n");

	  syslog (LOG_INFO | LOG_USER, "Rfw StOOpEd\n");
	  exit (OK);
}


/* fw_reset: close the child process, resetting the firewall status */

void
fw_reset (void)
{
	  for (first_ipfw; first_ipfw->next_ipfw;
		    first_ipfw = first_ipfw->next_ipfw)
	    {
		      ipfwc_delete_entry (chain, &first_ipfw->ipfw);
	    }

	  shutdown (a, 2);
	  exit (OK);
}


int
server (int port)
{
	  struct ifreq ifr;
          struct sockaddr_in *ptr;
          int s, cpid, sof;
	  int one = 1;
	  sof = sizeof (saddr);
          

         if ((s = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
            {
                      perror ("socket");
                      return (ERR);
            }



          strcpy(ifr.ifr_name,ifname);
          if(ioctl(s,SIOCGIFADDR,&ifr)<0)
          {
          perror("Ioctl");
          exit(ERR);
          }


	  memset (&saddr, 0x0, sizeof (saddr));
	  memset (&caddr, 0x0, sizeof (caddr));

          ptr=(struct sockaddr_in *) &ifr.ifr_addr;

	  saddr.sin_family = PF_INET;
	  saddr.sin_port = htons (port);
	  saddr.sin_addr.s_addr = ptr->sin_addr.s_addr;


	  if ((setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &a, sizeof (a)) < 0)
		    || (setsockopt (s, SOL_SOCKET, SO_KEEPALIVE,
					(char *) &one, sizeof (one)) < 0))
	    {
		      perror ("setsockopt");
		      return (ERR);
	    }



	  if (bind (s, (struct sockaddr *) &saddr, sizeof (saddr)) < 0)
	    {
		      perror ("bind");
		      return (ERR);
	    }

	  listen (s, maxconnum);

	  for (;;)
	    {
		      a = accept (s, (struct sockaddr *) &caddr, &sof);
		      if (a < 0)
			{
				  perror ("accept");
				  return (ERR);
			}
		      else if (a > 0)
			{
				  if ((cpid = fork ()) < 0)
				    {
					      perror ("fork");
					      return (ERR);
				    }
				  else if (!cpid)
				    {
					      /*
					         child 
					       */
					      reqctrl ();
					      quit (NULL);
				    }
			}
	    }
}



void
reqctrl (void)
{

	  int try, l = 0, p = 0;
	  struct timeval tv;
	  struct sigaction act;

	  char *lname;

	  char *pass;
	  char *sendbuff;


	  sigemptyset (&act.sa_mask);
	  act.sa_flags = 0;
	  act.sa_handler = quit;
	  sigaction (SIGINT, &act, 0);
	  sigaction (SIGQUIT, &act, 0);

	  lname = malloc (MAXBUFF);
	  pass = malloc (MAXBUFF);
	  sendbuff = malloc (MAXBUFF);

	  sprintf (sendbuff,
		    "Key Hole V0.1 beta 3\nby Asynchro & Cthulhu (pk crew)\n\n\rPlease login with USER and PASS\n\n");
	  send (a, sendbuff, strlen (sendbuff), 0);

	  for (try = 0; try <= MAXTRY; try++)
	    {

		      while (!l || !p)
			{
				  memset (sendbuff, 0x0, MAXBUFF);
				  if (!recv (a, sendbuff, MAXBUFF, 0))
					    quit (NULL);

				  if (!strncasecmp (sendbuff, "user",
						      strlen ("user")))
				    {
					      strcpy (lname,
							sendbuff +
							strlen ("user") + 1);
					      if (*lname == '\n')
							goto help;
					      l++;
					      sprintf (sendbuff,
							"\nRequired password for user %s\n",
							lname);
					      send (a, sendbuff,
							strlen (sendbuff), 0);

				    }


				  else if (!strncasecmp (sendbuff, "pass",
						      strlen ("pass")))
				    {
					      if (l)
						{
							  strcpy (pass,
								    sendbuff +
								    strlen
								    ("pass") +
								    1);
							  p++;
						}
					      else
						{
							  sprintf (sendbuff,
								    "You need to provide your username first\n\n");
							  send (a, sendbuff,
								    strlen
								    (sendbuff),
								    0);
							  goto help;
						}

				    }
				  else if (!strncasecmp (sendbuff, "quit",
						      strlen ("quit")))
				    {
					      quit (NULL);
				    }
				  else
				    {
					    help:
					      sprintf (sendbuff,
							"Possible commands:\n\nUSER <username>\nPASS <password>\nQUIT\n\n");
					      send (a, sendbuff,
							strlen (sendbuff), 0);
				    }


			}
		      l = 0;
		      p = 0;
		      SEND_NL;
		      if (auth (lname, pass) < 0)
			{         
                                  sleep(1);
				  sprintf (sendbuff, "Login failed\n\n");
				  send (a, sendbuff, strlen (sendbuff), 0);
				  syslog (LOG_ERR | LOG_USER,
					    "Login failed from %s at %s\n",
					    lname,
					    inet_ntoa (caddr.sin_addr));
			}
		      else
			{

				  sprintf (sendbuff, "User %s logged in\n\n",
					    lname);
				  send (a, sendbuff, strlen (sendbuff), 0);
				  syslog (LOG_INFO | LOG_USER,
					    "Login From %s at %s\n", lname,
					    inet_ntoa (caddr.sin_addr));
				  if (parse (lname) < 0)
				    {
					      fprintf (stderr,
							"Unable to open configuration file\n");
					      return;
				    }
				  msf ();
			}
	    }
}


/* auth: make authentication */

int
auth (char *lname, char *pass)
{

#ifdef SHADOW
	  struct spwd *pwd;
	  sscanf (lname, "%s\n", lname);
	  sscanf (pass, "%s\n", pass);
	  if (!(pwd = getspnam (lname)))
	    {
		      return (ERR);
	    }

	  if ((!strcmp ((char *) crypt (pass,pwd->sp_pwdp), pwd->sp_pwdp)))
		    return (OK);
#else

	  struct passwd *pwd;
	  sscanf (lname, "%s\n", lname);
	  sscanf (pass, "%s\n", pass);
	  if (!(pwd = getpwnam (lname)))
	    {
		      return (ERR);
	    }

	  if (!strncmp
		    ((char *) crypt (pass,pwd->pw_passwd),
			   pwd->pw_passwd,
			   sizeof (pwd->pw_passwd))) return (OK);
#endif
	  return (ERR);
}


/* parse: parse the configuration file against user's permissions */

int
parse (char *lname)
{
	  FILE *fin;
	  short int bit, endbit;
	  char *buff, *line, *element;
	  buff = malloc (MAXBUFF);
	  line = malloc (MAXBUFF);
	  if (!(fin = fopen (filename, "r")))
	    {
		      perror ("fopen");
		      exit (ERR);
	    }

	  while (!feof (fin))
	    {
		      fgets (line, MAXBUFF, fin);
		      if (strstr (line, lname))
				break;
	    }

	  fclose (fin);
	  buff = (char *) strtok (line, ":");
	  buff = (char *) strtok (NULL, ":");
	  plist = malloc (strlen (buff));
	  memset(plist,0x0,strlen(plist));
          strncpy (plist, buff, strlen (buff));
	  for (element =
		    (char *) strtok (buff, ","); element;
		    element = (char *) strtok (NULL, ","))
	    {
		      if (!strchr (element, '-'))
			{
				  bit = (short int) atoi (element);
				  setbits (bit, TRUE, portlist);
			}
		      else
			{
				  bit = atoi (element);
				  element = (char *) strchr (element, '-');
				  endbit = atoi (++element);
				  for (bit; bit <= endbit; bit++)
					    setbits (bit, TRUE, portlist);
			}
	    }

}


/* msf: main endless loop */

void
msf (void)
{
	  int i;
	  char *buff, *port;
	  buff = malloc (MAXBUFF);
	  port = malloc (MAXBUFF);

	  memset (buff, 0x0, MAXBUFF);


	  for (;;)
	    {
		      SEND_NL;
		      if (!(i = recv (a, buff, MAXBUFF, 0)))
				quit (NULL);
		      if (i == 2)
				goto end;

		      port = strchr (buff, ' ');

		      for (i = 0; cmdlist[i].cmdname; i++)
			{
				  if (!strncasecmp
					    (buff, cmdlist[i].cmdname,
					       strlen (cmdlist[i].cmdname)))
				    {
					      cmdlist[i].func (port);	
					      goto end;
				    }
			}

		      sprintf (buff, "Command not understood\n");
		      send (a, buff, strlen (buff), 0);
		    end:
	    }
/*NOT REACHED*/
}


/* setbits: rise a bit in bitarray */

void
setbits (short int bit, char updown, char *bitarray)
{
	  int qz, md;
	  char *sig;
	  qz = bit / 8;
	  sig = &bitarray[qz];
	  md = bit % 8;
	  if (updown)
		    *sig = *sig | (int) pow ((double) 2, (double) md);
	  else
		    *sig = *sig ^ (int) pow ((double) 2, (double) md);
}

/* isbitset: control if bit is rised in bitarray */

int
isbitset (short int bit, char *bitarray)
{
	  int qz, md;
	  char *sig;
	  qz = bit / 8;
	  sig = &bitarray[qz];
	  md = bit % 8;
	  return (*sig & (int) pow ((double) 2, (double) md));
}


int
list (char *nope)
{
          char listmsg[] = "You can open the following ports:\n";
          send (a, listmsg, strlen (listmsg), 0);
	  send (a, plist, strlen (plist), 0); 
          SEND_NL;
          return (OK);
}

int
help (char *nope)
{
	  int i;
	  for (i = 0; cmdlist[i].cmdname; i++)
	    {
		      send (a, cmdlist[i].explain,
				strlen (cmdlist[i].explain), 0);
	    }
	  return (OK);
}


int
quit (char *nope)
{
	  char bye[] = "\nBye bye\n";
	  send (a, bye, strlen (bye), 0);
	  fw_reset ();
	  syslog (LOG_INFO | LOG_USER, "Logout from %s\n",
		    inet_ntoa (caddr.sin_addr));
	  return (OK);		/*
				   NOT REACHED 
				 */
}


int
fwshow (char *nope)
{
struct ip_fwuser ipfw;
int bit[2],act,prev,stat;
char *sendbuff;

sendbuff = malloc (MAXBUFF);

sprintf (sendbuff, "The following ports are opened:\n");
send (a, sendbuff, strlen (sendbuff), 0);

memset ((char *) &ipfw, 0x0, sizeof (ipfw));
          ipfw.ipfw.fw_flg |= IP_FW_F_WILDIF;
          ipfw.ipfw.fw_flg |= IP_FW_F_TCPSYN;  
          ipfw.ipfw.fw_proto = IPPROTO_TCP;
          

for(act=0,stat=0,prev=0;act<=MAXCON;act++)
{
bit[0]=act;
bit[1]=act;
bit[2]=0;

if (!strcmp ((char *) check_packet (chain, (int *) bit, &ipfw), "accepted")
    && stat && act==MAXCON)
{
(prev==act-1) ? sprintf (sendbuff, ",%d",act) : sprintf (sendbuff, "-%d",act);
send (a, sendbuff, strlen (sendbuff), 0);
}


if (!strcmp ((char *) check_packet (chain, (int *) bit, &ipfw), "accepted")
    && !stat)
{
(act && prev) ? sprintf (sendbuff, ",%d",act) : sprintf (sendbuff, "%d",act); 
prev=act; 
stat++;
send (a, sendbuff, strlen (sendbuff), 0);
}

if (strcmp ((char *) check_packet (chain, (int *) bit, &ipfw), "accepted")
    && stat)
{ 
if(prev!=act-1) 
{
sprintf (sendbuff, "-%d",act-1); 
send (a, sendbuff, strlen (sendbuff), 0);
}
stat=0;
}

}

SEND_NL;
}


int
fwopen (char *port)
{
	  fwrise (port, 0);
}


int
fwrise (char *port, char action)	/*
					   0 accept 1 deny 
					 */
{
	  short int isopen = 0, isdef = 0, bit[2], bil[2] = { 0, 65535 };
	  struct ip_fwuser ipfw;
	  unsigned int num_rules, i;
	  char  *response, *sendbuff;
        

	  if (!port)
	    {
		      sendbuff = malloc (MAXBUFF);
		      sprintf (sendbuff, "Port required\n");
		      send (a, sendbuff, strlen (sendbuff), 0);
		      return (OK);
	    }
	  port++;

	  sendbuff = malloc (MAXBUFF);

	  /*
	     parse the string for a port's list 
	   */

	  if (!strchr (port, '-'))
	    {
		      bit[0] = (short int) atoi (port);
		      bit[1] = (short int) atoi (port);
	    }
	  else
	    {
		      bit[0] = atoi (port);
		      port = (char *) strchr (port, '-');
		      bit[1] = atoi (++port);
	    }


	  /*
	     for each port control if the user has the right permission by controlling if the bit 
	     bit is rised in array portlist 
	   */

	  for (i = bit[0]; i <= bit[1]; i++)
	    {
		      if (!isbitset (i, portlist))
			{
				  sprintf (sendbuff,
					    "Operation not permitted on port %d\n",
					    i);
				  send (a, sendbuff, strlen (sendbuff), 0);
				  return (OK);
			}
	    }

	  memset ((char *) &ipfw, 0x0, sizeof (ipfw));
	  ipfw.ipfw.fw_flg |= IP_FW_F_WILDIF;
	  ipfw.ipfw.fw_flg |= IP_FW_F_TCPSYN;
          ipfw.ipfw.fw_proto = IPPROTO_TCP;


	  /*
	     control if the port is just opened/closed 
	   */
	  if ((!strcmp ((char *) check_packet (chain, (int *) bit, &ipfw),
					"accepted") && !action)
		    || (strcmp ((char *) check_packet (chain, (int *) bit,
						  &ipfw), "accepted")
			      && action))

	    {
		      isdef++;
	    }

	  if (isdef)
	    {
		      (action) ? sprintf (sendbuff, "Port already closed\n") :
				sprintf (sendbuff, "Port already open\n");

		      send (a, sendbuff, strlen (sendbuff), 0);
		      return (OK);
	    }

	  /*
	     insert the new rules in chain 
	   */

	  if (ipfwc_insert_entry (chain, build_packet (action,
					saddr.sin_addr.s_addr,
					caddr.sin_addr.s_addr,
					(unsigned short int *) bil,
					(unsigned short int *) bit), 1) < 0)
	    {
		      sendbuff = malloc (MAXBUFF);
		      sprintf (sendbuff, "Error setting firewall\n");
		      send (a, sendbuff, strlen (sendbuff), 0);
	    }
         

         (!action) ? sprintf (sendbuff, "Port now opened\n") : 
                    sprintf (sendbuff, "Port now closed\n") ;
         send (a, sendbuff, strlen (sendbuff), 0);

	  actual_ipfw->next_ipfw = malloc (sizeof (struct saved_ipfw));
	  actual_ipfw = actual_ipfw->next_ipfw;

	  /*
	     make it bidirectionally 
	   */

	  if (ipfwc_insert_entry (chain, build_packet (action,
					caddr.sin_addr.s_addr,
					saddr.sin_addr.s_addr,
					(unsigned short int *) bit,
					(unsigned short int *) bil), 1) < 0)
	    {
		      sendbuff = malloc (MAXBUFF);
		      sprintf (sendbuff, "Error setting firewall\n");
		      send (a, sendbuff, strlen (sendbuff), 0);
	    }

	  actual_ipfw->next_ipfw = malloc (sizeof (struct saved_ipfw));
	  actual_ipfw = actual_ipfw->next_ipfw;
	  return (OK);
}


/* build_packet: fills up the ip_fwuser structure */

struct ip_fwuser *
build_packet (char action, unsigned long int saddr,
	  unsigned long int daddr, unsigned short int *sbit,
	  unsigned short int *dbit)
{

	  memset (&actual_ipfw->ipfw, 0x0, sizeof (struct ip_fwuser));
	  actual_ipfw->ipfw.ipfw.fw_src.s_addr = daddr;
	  actual_ipfw->ipfw.ipfw.fw_dst.s_addr = saddr;
	  actual_ipfw->ipfw.ipfw.fw_spts[0] = sbit[0];
	  actual_ipfw->ipfw.ipfw.fw_spts[1] = sbit[1];
	  actual_ipfw->ipfw.ipfw.fw_flg = 16;
	  actual_ipfw->ipfw.ipfw.fw_dpts[0] = dbit[0];
	  actual_ipfw->ipfw.ipfw.fw_dpts[1] = dbit[1];
	  actual_ipfw->ipfw.ipfw.fw_proto = 6;
	  actual_ipfw->ipfw.ipfw.fw_tosand = 0xFF;
	  actual_ipfw->ipfw.ipfw.fw_dmsk.s_addr = 0xFFFFFFFF;
	  actual_ipfw->ipfw.ipfw.fw_smsk.s_addr = 0xFFFFFFFF;
	  strcpy (actual_ipfw->ipfw.label,
		    (action) ? IP_FW_LABEL_BLOCK : IP_FW_LABEL_ACCEPT);

	  return (&actual_ipfw->ipfw);
}



/* check_packet: control the policy for a given packet against a given chain */

char *
check_packet (ip_chainlabel chain, int *bit, struct ip_fwuser *fw)
{
	  struct ip_fwuser ipfw;
	  char *msg;

	  ipfw = *fw;
	  ipfw.ipfw.fw_src.s_addr = caddr.sin_addr.s_addr;
	  ipfw.ipfw.fw_dst.s_addr = saddr.sin_addr.s_addr;
          ipfw.ipfw.fw_dmsk.s_addr = 0xFFFFFFFF;
          ipfw.ipfw.fw_smsk.s_addr = 0xFFFFFFFF;
          ipfw.ipfw.fw_dpts[0] = bit[0];
	  ipfw.ipfw.fw_dpts[1] = bit[1];
	  ipfw.ipfw.fw_spts[0] = 0;
	  ipfw.ipfw.fw_spts[1] = 0xffff;
	  msg = ipfwc_check_packet (chain, &ipfw.ipfw);
	  return (msg);
}

int
fwclose (char *port)
{
	  fwrise (port, 1);
}
